/*
  file:    nnhood.c
  purpose: built tcp/ip "network neighborhood" using UDP echo broadcast
  created: pasha nov 13-24 1998
  notes:   1. seems that a standard win9x machine does not answer
           on echo port, so they will be not included, but who cares
           about this crap? nt should be ok as long as what is called
	   there "simple TCP/IP services" is installed; to use with
           unices, be sure the UDP echo string is not commented out
           in inetd.conf
	   2. works on Linux, SunOS, Solaris and HP-UX.
*/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if.h>


#define UDP_PORT 7        /* echo port */
#define BROADCAST_MESSAGE "hey you little bastard"
#define IFCONF_BUFFER_LENGTH 4096      /* length of buffer to store 
                                          interfaces list */

void gasp (const char*);
struct in_addr* check_interface (struct ifreq);

static char error_string[80];
static int SOCK;


int main (int argc, char* argv[])
{
   int rval, timeout, i;
   int sockopt_val = 1;
   int len_broad_msg = strlen(BROADCAST_MESSAGE);    
   char recvline[len_broad_msg+1];       /* this possibly will 
					    not work on not gcc */
         
   struct sockaddr_in from_addr, to_addr;
   
   struct sockaddr rec_addr;
   int len_rec_addr = sizeof(rec_addr);
   struct in_addr rec_in_addr;
   char rec_in_addr_dot_alloc[16];      /* place for "dotted" ip address;
				          enough as long as we are not 
				          in iv6 world :-) */
   char* rec_in_addr_dot = rec_in_addr_dot_alloc;
   struct hostent rec_hostent_alloc;
   struct hostent* rec_hostent = &rec_hostent_alloc;
   
   struct in_addr broad_in_addr_alloc;  /* place for broadcast address */
   struct in_addr* broad_in_addr = &broad_in_addr_alloc;
   
   /* select-related */
   fd_set rfds;
   struct timeval tv;
   
   struct ifconf if_conf;           /* "interfaces list" struct for ioctl */
   struct ifreq* if_req;            /* "interface request" struct for ioctl */
   char buf[IFCONF_BUFFER_LENGTH];  /* buffer to store "interfaces list" */
   
   if ((argc != 2) && (argc != 3))
     {
	fprintf (stderr, "Usage: %s <timout_in_sec> [network_interface]\n", 
		 argv[0]);
	exit (0);
     }

   /* set timeout in tv struct used in select() */
   timeout = atoi(argv[1]);
   tv.tv_sec = timeout;
   tv.tv_usec = 0;
   
   /* fill-in addresses structs */
   memset ((void*)&from_addr, 0, sizeof(from_addr));
   from_addr.sin_family = AF_INET;
   from_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	
   memset ((void*)&to_addr, 0, sizeof(to_addr));
   to_addr.sin_family = AF_INET;
   /* to_addr.sin_addr.s_addr = htonl(INADDR_BROADCAST); */
   /*
    on some systems (e.g. Linux Debian 1.2 and SunOS) an attempt to broadcast
    when binding INADDR_BROADCADT address (which is 255.255.255.255) 
    results in error "network in unreachable" from sendto();
    I don't know whether this is system bug/misconfiguration,
    or a feature; also, some (recent) postings of Rich Stevens 
    make me think that this 255.255.255.255 thingy is obsolete
    which will not necessary be supported in the future;
    so we trying to find below a real broadcast address 
    (as e.g. reported by netconfig) and bind it instead 
    as this works in all cases; this also probably the only right 
    thing to do in the case of several network interfaces
    */
      
   to_addr.sin_port = htons(UDP_PORT);


   /* create socket */
   if ((SOCK = socket (PF_INET, SOCK_DGRAM, IPPROTO_IP)) < 0)
     { gasp ("socket"); }


   /* get interfaces list */
   if_conf.ifc_len = sizeof(buf);
   if_conf.ifc_buf = buf;
   if (ioctl (SOCK, SIOCGIFCONF, &if_conf) != 0)
     { gasp ("ioctl SIOCGIFCONF"); }

   /* loop over all interfaces */
   if_req = if_conf.ifc_req;
   for (i = (if_conf.ifc_len)/sizeof(struct ifreq); --i >= 0; if_req++)
     {
	/* if there was second argument, it is interface name
	 * and then we check whether current interface coincides with it
	 */
	if ((argc != 3) || (strcmp (if_req->ifr_name, argv[2]) == 0))
	  {
	     if ((broad_in_addr = check_interface (*if_req)) != NULL)
	       {
		  to_addr.sin_addr.s_addr = broad_in_addr->s_addr;
		  fprintf (stderr, "using interface %s with broadcast address %s\n", 
			   if_req->ifr_name, inet_ntoa(*broad_in_addr));
		  break;
	       }
	     if (argc == 3)
	       {
		  sprintf (error_string, "interface %s is not suitable", 
			   argv[2]);
		  gasp (error_string);
	       }
	  }
     }
   
   if (to_addr.sin_addr.s_addr == 0)
     {
	if (argc == 3)
	  {
	     sprintf (error_string, "interface %s does not exists", argv[2]);
	     gasp (error_string);
	  }
	else
	  { gasp ("no suitable interface found"); }
     }
   
   
   if (setsockopt (SOCK, SOL_SOCKET, SO_BROADCAST, (char*)&sockopt_val, sizeof(sockopt_val)) < 0)
     { gasp ("setsocketopt"); }
   /* should we also set SO_REUSEADDR option? seems no */
   
   if (bind (SOCK, (struct sockaddr*)&from_addr, sizeof(from_addr)) < 0)
     { gasp ("bind"); }
  
   if (sendto (SOCK, BROADCAST_MESSAGE, len_broad_msg, 0, 
	       (struct sockaddr*)&to_addr, sizeof(to_addr)) != len_broad_msg)
     { gasp ("sendto"); }

   FD_ZERO (&rfds);
   FD_SET (SOCK, &rfds);
   
   while ((rval = select (SOCK+1, &rfds, NULL, NULL, &tv)) == 1)
     {
	/* socket is ready for reading */
	if (recvfrom (SOCK, recvline, len_broad_msg, 0, &rec_addr, 
		      &len_rec_addr) != len_broad_msg)
	  { gasp ("recvfrom"); }
		
	rec_in_addr.s_addr = 
	  (((struct sockaddr_in*)&rec_addr)->sin_addr).s_addr;
	rec_in_addr_dot = inet_ntoa (rec_in_addr);
		
	printf ("%s", rec_in_addr_dot);
	if ((rec_hostent = gethostbyaddr((char*)&rec_in_addr, sizeof(rec_in_addr), AF_INET)) 
	    == NULL)
	  { fprintf (stderr, "host information for %s not found\n", rec_in_addr_dot); }
	else
	  { printf (" %s", rec_hostent->h_name); }
	printf ("\n");

	if (strcmp (recvline, BROADCAST_MESSAGE) != 0)
	  { fprintf (stderr, "wrong answer %s from %s\n", recvline, rec_in_addr_dot); }
     }
   
   if (rval == -1)
     {
	/* error in select occured */
	perror ("select");
     }
   else
     {
	/* timeout expired */
	fprintf (stderr, "timeout of %d second(s) expired\n", timeout);
     }
         
   close (SOCK);
   exit (0);
}



/* checks given network interface 
 * input parm - ifreq structure
 * return: in_addr structure of broadcast address
 * if this interface is "suitable", NULL otherwise or if some error occurs
 * we assume that SOCK is already opened
 */
struct in_addr* check_interface (struct ifreq if_req)
{
   /* get interface flags */
   if (ioctl (SOCK, SIOCGIFFLAGS, &if_req) != 0)
     {
	sprintf (error_string, "ioctl SIOCGIFFLAGS on %s", 
		 if_req.ifr_name); 
	perror (error_string);
	return (NULL);
     }
   
   /* check that this interface is up, supports broadcast,
      not loopback and resources-allocated */
   if ( (if_req.ifr_flags & IFF_UP) &&
       (if_req.ifr_flags & IFF_BROADCAST) &&
       !(if_req.ifr_flags & IFF_LOOPBACK) &&
       (if_req.ifr_flags & IFF_RUNNING)
       )
     {
	/* get broadcast address */
	if (ioctl (SOCK, SIOCGIFBRDADDR, &if_req) != 0) 
	  { 
	     sprintf (error_string, "ioctl SIOCGIFBRDADDR on %s",
		      if_req.ifr_name);
	     perror (error_string);
	     return (NULL);
	  }
	return (&((struct sockaddr_in*)&(if_req.ifr_broadaddr))->sin_addr);
     }
   else 
     { return (NULL); }
}




void gasp (const char* msg)
{
   perror (msg);
   close (SOCK);
   exit (-1);
}

